# -*- coding: utf-8 -*-
import tempfile

from mock import patch

from xprofile.xrandr import Xrandr
from xprofile.__main__ import main


@patch('xprofile.xrandr.Popen')
def test_create_new_profile(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    with tempfile.NamedTemporaryFile() as tmpfile:
        Popen.return_value.communicate.return_value = (xrandr_stdout, None)
        Popen.return_value.wait.return_value = 0

        retval = main(['--config', tmpfile.name, 'create', 'testprofile'])

        assert retval == 0
        assert Popen.called
        assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']


@patch('xprofile.xrandr.Popen')
def test_create_new_profile_dryrun(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    with tempfile.NamedTemporaryFile() as tmpfile:
        Popen.return_value.communicate.return_value = (xrandr_stdout, None)
        Popen.return_value.wait.return_value = 0

        retval = main(['--config', tmpfile.name, 'create', 'testprofile', '--dry-run'])

        assert retval == 0
        assert Popen.called
        assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']


@patch('xprofile.xrandr.Popen')
def test_create_existing_profile(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'create', 'laptop'])

    assert retval == 1
    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']

def test_list():
    main(['--config', 'test/xprofilerc_example', 'list'])


@patch('xprofile.xrandr.Popen')
def test_current(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'current'])

    assert retval == 0
    assert Popen.called


@patch('xprofile.xrandr.Popen')
def test_activate_profile(Popen):
    Popen.return_value.communicate.return_value = (b'', None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'activate', 'docked'])

    assert retval == 0
    assert Popen.called
    assert Popen.call_args[0][0] == [
        '/usr/bin/xrandr',
        '--output', 'LVDS1', '--off',
        '--output', 'DP2', '--mode', '1920x1080', '--pos', '0x500', '--primary',
        '--output', 'HDMI3', '--mode', '1920x1080', '--rotate', 'left', '--pos', '1930x0'
    ]


def test_activate_profile_nonexistent():
    retval = main(['--config', 'test/xprofilerc_example', 'activate', 'nonexistent'])

    assert retval == 1


@patch('xprofile.xrandr.Popen')
def test_activate_profile_auto(Popen):
    with open('test/laptop.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'auto'])

    assert retval == 0
    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--auto']


@patch('xprofile.xrandr.Popen')
def test_activate_profile_auto__dryrun(Popen):
    with open('test/laptop.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'auto', '--dry-run'])

    assert retval == 0
    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']


@patch('xprofile.xrandr.Popen')
def test_activate_profile_auto_nonexistent(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    retval = main(['--config', 'test/xprofilerc_example', 'auto'])

    assert retval == 0
    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--auto']



# -*- coding: utf-8 -*-
from mock import patch
from xprofile.xrandr import Xrandr


@patch('xprofile.xrandr.Popen')
def test_call_xrandr_failure(Popen):
    Popen.return_value.communicate.return_value = (None, 'An unknown error occurred.')
    Popen.return_value.wait.return_value = 1

    try:
        edid = Xrandr().get_screen()
    except RuntimeError as err:
        assert str(err) == 'xrandr error: An unknown error occurred.'
    else:
        assert False, 'Failed to raise RuntimeError'

    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']


@patch('xprofile.xrandr.Popen')
def test_call_xrandr_set_display(Popen):
    Popen.return_value.communicate.return_value = (b'', None)
    Popen.return_value.wait.return_value = 0

    Xrandr(display=':1').call_xrandr()

    assert Popen.called
    assert Popen.call_args[1]['env']['DISPLAY'] == ':1'


@patch('xprofile.xrandr.Popen')
def test_xrandr_get_screen_multiple(Popen):
    with open('test/multi_screen.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    screen = Xrandr().get_screen()

    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']
    assert len(screen['displays']) == 3

    assert screen['displays'][0]['name'] == 'VGA1'
    assert screen['displays'][0]['connected'] == True
    assert screen['displays'][0]['status'] == 'connected'
    assert screen['displays'][0]['geometry']['dimension'] == '1280x1024'
    assert screen['displays'][0]['geometry']['offset'] == '0x0'
    assert screen['displays'][0]['active'] == True
    assert screen['displays'][0]['rotation'] == None
    assert screen['displays'][0]['primary'] == False
    assert type(screen['displays'][0]['edid']) == list
    assert type(screen['displays'][0]['modes']) == dict

    assert screen['displays'][1]['name'] == 'DVI1'
    assert screen['displays'][1]['connected'] == True
    assert screen['displays'][1]['status'] == 'connected'
    assert screen['displays'][1]['geometry']['dimension'] == '1280x1024'
    assert screen['displays'][1]['geometry']['offset'] == '1280x0'
    assert screen['displays'][1]['active'] == True
    assert screen['displays'][1]['rotation'] == None
    assert screen['displays'][1]['primary'] == False
    assert type(screen['displays'][1]['edid']) == list
    assert type(screen['displays'][1]['modes']) == dict

    assert screen['displays'][2]['name'] == 'TV1'
    assert screen['displays'][2]['connected'] == False
    assert screen['displays'][2]['status'] == 'unknown connection'
    assert screen['displays'][2]['geometry'] == None
    assert type(screen['displays'][2]['edid']) == list
    assert type(screen['displays'][2]['modes']) == dict

    assert screen.get_xrandr_options() == [
        '--output', 'VGA1', '--mode', '1280x1024', '--pos', '0x0',
        '--output', 'DVI1', '--mode', '1280x1024', '--pos', '1280x0'
    ]

    assert screen.get_edid() == 'd41d8cd98f00b204e9800998ecf8427e'


@patch('xprofile.xrandr.Popen')
def test_xrandr_get_screen_docked(Popen):
    with open('test/docked.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    screen = Xrandr().get_screen()

    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']
    assert len(screen['displays']) == 8

    for index in [0, 5, 6]:
        assert screen['displays'][index]['connected'] == True
        assert screen['displays'][index]['status'] == 'connected'
        assert type(screen['displays'][index]['edid']) == list
        assert type(screen['displays'][index]['modes']) == dict

    for index in [1, 2, 3, 4, 7]:
        assert screen['displays'][index]['connected'] == False
        assert screen['displays'][index]['status'] == 'disconnected'
        assert screen['displays'][index]['geometry'] == None
        assert screen['displays'][index]['primary'] == False
        assert screen['displays'][index]['active'] == False
        assert screen['displays'][index]['mode'] == None
        assert type(screen['displays'][index]['edid']) == list
        assert type(screen['displays'][index]['modes']) == dict

    assert screen['displays'][0]['name'] == 'LVDS1'
    assert screen['displays'][0]['geometry'] == None
    assert screen['displays'][0]['mode'] == None
    assert '(0x4b)' in screen['displays'][0]['modes']

    assert screen['displays'][5]['name'] == 'HDMI3'
    assert screen['displays'][5]['geometry']['dimension'] == '1080x1920'
    assert screen['displays'][5]['geometry']['offset'] == '1930x0'
    assert screen['displays'][5]['rotation'] == 'left'
    assert screen['displays'][5]['mode'] in screen['displays'][5]['modes']
    assert screen['displays'][5]['mode'] == '(0xc9)'
    assert screen['displays'][5]['modes']['(0xc9)']['current'] == True

    assert screen['displays'][6]['name'] == 'DP2'
    assert screen['displays'][6]['geometry']['dimension'] == '1920x1080'
    assert screen['displays'][6]['geometry']['offset'] == '0x500'

    assert screen.get_xrandr_options() == [
        '--output', 'LVDS1', '--off',
        '--output', 'HDMI3', '--mode', '1920x1080', '--pos', '1930x0', '--rotate', 'left',
        '--output', 'DP2',   '--primary', '--mode', '1920x1080', '--pos', '0x500'
    ]

    assert screen.get_edid() == 'c2989146488f57fa9dc5f7efc263b0fd'


@patch('xprofile.xrandr.Popen')
def test_xrandr_get_screen_laptop(Popen):
    with open('test/laptop.txt', 'rb') as file:
        xrandr_stdout = file.read()

    Popen.return_value.communicate.return_value = (xrandr_stdout, None)
    Popen.return_value.wait.return_value = 0

    screen = Xrandr().get_screen()

    assert Popen.called
    assert Popen.call_args[0][0] == ['/usr/bin/xrandr', '--verbose']
    assert len(screen['displays']) == 8

    assert screen['displays'][0]['name'] == 'LVDS1'
    assert screen['displays'][0]['connected'] == True
    assert screen['displays'][0]['status'] == 'connected'
    assert screen['displays'][0]['mode'] == '(0x4b)'
    assert screen['displays'][0]['mode'] in screen['displays'][0]['modes']
    assert screen['displays'][0]['modes']['(0x4b)']['current'] == True
    assert screen['displays'][0]['geometry']['dimension'] == '1920x1080'
    assert screen['displays'][0]['geometry']['offset'] == '0x0'
    assert screen['displays'][0]['rotation'] == None

    assert type(screen['displays'][0]['edid']) == list
    assert type(screen['displays'][0]['modes']) == dict

    for display in screen['displays'][1:]:
        assert display['connected'] == False
        assert display['status'] == 'disconnected'
        assert display['geometry'] == None
        assert display['primary'] == False
        assert display['active'] == False
        assert display['mode'] == None
        assert type(display['edid']) == list
        assert type(display['modes']) == dict

    assert screen.get_xrandr_options() == [
        '--output', 'LVDS1', '--mode', '1920x1080', '--pos', '0x0'
    ]

    assert screen.get_edid() == 'cfdee1377d86e245f2d187082f7a504a'



# -*- coding: utf-8 -*-
import os
import logging

from re import compile
from hashlib import md5
from subprocess import Popen, PIPE


log = logging.getLogger(__name__)


RE_EDID = compile(r'^\s*([a-f0-9]{32})$')
RE_XRANDR_DISPLAY = compile(r'^(?P<name>[^\s]+) (?P<status>disconnected|connected|unknown connection) ?(?P<primary>primary)? ?(?P<geometry>(?P<width>[0-9]+)x(?P<height>[0-9]+)\+(?P<x>[0-9]+)\+(?P<y>[0-9]+))? ?(?P<mode>\([0-9a-fx]+\))? ?(?P<rotation>normal|left|inverted|right)?.*$')
RE_XRANDR_SCREEN = compile(r'^Screen (?P<screen>[0-9]+): minimum (?P<minimum>[0-9]+ x [0-9]+), current (?P<current>[0-9]+ x [0-9]+), maximum (?P<maximum>[0-9]+ x [0-9]+)')
RE_DISPLAY_MODE = compile(r'^\s+(?P<dimension>[0-9x]+) (?P<modeid>\([0-9a-fx]+\)) [0-9\.]+MHz [-+]HSync [-+]VSync ?(?P<current>\*current)? ?(?P<preferred>\+preferred)?\s*$')


class Screen(dict):
    def __init__(self, *args, **kwargs):
        self['displays'] = []
        super(Screen, self).__init__(*args, **kwargs)

    def get_xrandr_options(self):
        '''
        Loop through all active displays and create an xrandr compatible
        option list.
        '''
        line = []
        for display in self['displays']:
            line += display.get_xrandr_options()
        return line

    def get_edid(self):
        '''
        Get a md5 hash of all EDID of all currently connected displays
        '''
        md5sum = md5()

        for display in self['displays']:
            for edid in display['edid']:
                md5sum.update(edid)

        return md5sum.hexdigest()


class Display(dict):
    def __init__(self, *args, **kwargs):
        self['edid'] = []
        self['modes'] = {}
        self['geometry'] = None
        super(Display, self).__init__(*args, **kwargs)

    def get_xrandr_options(self):
        if not self['connected']:
            return []

        line = ['--output', self['name']]

        if not self['active']:
            line += ['--off']

            return line

        if self['primary']:
            line += ['--primary']

        if self['mode'] in self['modes']:
            mode = self['modes'][self['mode']]['dimension']
        else:
            # TODO: is this fallback needed?
            mode = self['geometry']['dimension']

        line += [
            '--mode', mode,
            '--pos',  self['geometry']['offset']
        ]

        if self['rotation']:
            line += ['--rotate', self['rotation']]

        return line


class Xrandr(object):
    def __init__(self, xrandr_bin='/usr/bin/xrandr', display=None):
        self.xrandr_bin = xrandr_bin
        self.display = display

    def get_screen(self):
        screen  = Screen()

        for line in self.call_xrandr(['--verbose']):
            parts = RE_XRANDR_DISPLAY.match(line)
            if parts:
                display = Display()
                display['name'] = parts.group('name')
                display['status'] = parts.group('status')
                display['connected'] = parts.group('status') == 'connected'
                display['primary'] = parts.group('primary') == 'primary'
                display['mode'] = parts.group('mode')

                if parts.group('rotation') == 'normal':
                    display['rotation'] = None
                else:
                    display['rotation'] = parts.group('rotation')

                if parts.group('geometry'):
                    display['geometry'] = {
                        'dimension': '%sx%s' % (parts.group('width'), parts.group('height')),
                        'offset':    '%sx%s' % (parts.group('x'), parts.group('y'))
                    }

                display['active'] = isinstance(display['geometry'], dict)
                screen['displays'].append(display)
                continue

            parts = RE_EDID.match(line)
            if parts:
                edid = line.strip().encode()
                screen['displays'][-1]['edid'].append(edid)
                continue

            parts = RE_DISPLAY_MODE.match(line)
            if parts:
                screen['displays'][-1]['modes'][parts.group('modeid')] = {
                    'id': parts.group('modeid'),
                    'dimension': parts.group('dimension'),
                    'current': parts.group('current') != None,
                    'preferred': parts.group('preferred') != None,
                }
                continue

        return screen

    def call_xrandr(self, args=[]):
        '''
        Make a call to the xrandr binary in a subprocess
        '''
        current_env = os.environ.copy()

        if self.display:
            current_env['DISPLAY'] = self.display

        process = Popen([self.xrandr_bin] + args, env=current_env,
                        stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        status = process.wait()

        if status != 0:
            message = 'xrandr error: {0}'.format(stderr)
            log.error(message)
            raise RuntimeError(message)

        return stdout.decode().split('\n')



# -*- coding: utf-8 -*-
import re
import codecs

from setuptools import setup
from setuptools.command.test import test as TestCommand


class NoseTestCommand(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # Run nose ensuring that argv simulates running nosetests directly
        import nose
        nose.run_exit(argv=['nosetests'])


setup(
    name = 'xprofile',
    description = 'A tool to manage and automatically apply xrandr configurations.',
    version = re.search(r'''^__version__\s*=\s*["'](.*)["']''', open('xprofile/__init__.py').read(), re.M).group(1),
    author = 'Nico Di Rocco',
    author_email = 'dirocco.nico@gmail.com',
    url = 'https://github.com/nrocco/xprofile',
    license = 'GPLv3',
    long_description = codecs.open('README.rst', 'rb', 'utf-8').read(),
    test_suite='nose.collector',
    install_requires = [
        'docutils>=0.12'
    ],
    tests_require = [
        'nose',
        'mock',
        'coverage',
    ],
    packages = [
        'xprofile'
    ],
    entry_points = {
        'console_scripts': [
            'xprofile = xprofile.__main__:main'
        ]
    },
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: X11 Applications',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Desktop Environment',
        'Topic :: Utilities'
    ],
    cmdclass = {
        'test': NoseTestCommand
    }
)
